/* ------------------------------------------------------------ */
/*
HTTrack Website Copier, Offline Browser for Windows and Unix
Copyright (C) Xavier Roche, Yann Philippot

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


Important notes:

- We hereby ask people using this source NOT to use it in purpose of grabbing
emails addresses, or collecting any other private information on persons.
This would disgrace our work, and spoil the many hours we spent on it.


This project has been developed by Xavier Roche and Yann Philippot,
from the company Serianet at Caen, France (http://www.serianet.com)

Please visit our Website: http://www.httrack.com
*/


/* ------------------------------------------------------------ */
/* File: Main file .h                                           */
/* Author: Xavier Roche                                         */
/* ------------------------------------------------------------ */

// Fichier librairie .h
#ifndef HTTRACK_DEFH
#define HTTRACK_DEFH


#include "htsglobal.h"

/* specific definitions */
#include "htsbase.h"
// Includes & dfinitions
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef _WIN32
#include <conio.h>
#include <signal.h>
#include <direct.h>
#else
#include <signal.h>
#include <unistd.h>
#endif
/* END specific definitions */


// Include htslib.h for all types
#include "htslib.h"

#include "htsopt.h"

// structure d'un lien
typedef struct {
  char firstblock;      // flag 1=premier malloc 
  char link_import;     // lien import  la suite d'un moved - ne pas appliquer les rgles classiques up/down
  short int prio;       // priorit du lien ; >0 forte 0=faible
  short int pass2;      // traiter aprs les autres, seconde passe
  int premier;          // pointeur sur le premier lien qui a donn lieu aux autres liens du domaine
  int precedent;        // pointeur sur le lien qui a donn lieu  ce lien prcis
  //int moved;            // pointeur sur moved
  short int retry;      // nombre de retry restants
  short int testmode;   // mode test uniquement, envoyer juste un head!
  char* adr;            // adresse
  char* fil;            // nom du fichier distant
  char* sav;            // nom  sauver sur disque (avec chemin ventuel)
  char* cod;            // chemin codebase ventuel si classe java
  char* former_adr;     // adresse initiale (avant ventuel moved), peut tre nulle
  char* former_fil;     // nom du fichier distant initial (avant ventuel moved), peut tre nul
  // pour optimisation:
#if HTS_HASH
  int hash_next[3];     // prochain lien avec mme valeur hash
#else
  int sav_len;          // taille de sav
#endif
} lien_url;

// chargement de fichiers en 'arrire plan'
typedef struct {
#if DEBUG_CHECKINT
  char magic;
#endif
  int status;           // status (-1=non utilis, 0: prt, >0: opration en cours)
  int testmode;         // mode de test
  char url_adr[HTS_URLMAXSIZE*2];     // adresse
  char url_fil[HTS_URLMAXSIZE*2];     // nom du fichier distant
  char url_sav[HTS_URLMAXSIZE*2];     // nom  sauver sur disque (avec chemin ventuel)
  char referer_adr[HTS_URLMAXSIZE*2]; // adresse host page referer
  char referer_fil[HTS_URLMAXSIZE*2]; // fichier page referer
  int timeout;            // grer des timeouts? (!=0  : nombre de secondes)
  double timeout_refresh; // si oui, time refresh
  int rateout;            // timeout refresh? (!=0 : taux minimum tolr en octets/s)
  double rateout_time;    // si oui, date de dpart
  LLint maxfile_nonhtml;  // taille max d'un fichier non html
  LLint maxfile_html;     // idem pour un ficheir html
  char location_buffer[HTS_URLMAXSIZE*2];  // "location" en cas de "moved" (302,..)
  htsblk r;               // structure htsblk de chaque objet en background 
  char send_too[1024];    // donnes  envoyer en mme temps que le header
  int head_request;       // requte HEAD?
  LLint range_req_size;   // range utilis
  //
  int http11;             // L'en tte doit tre sign HTTP/1.1 et non HTTP/1.0
  int is_chunk;           // chunk?
  char* chunk_adr;        // adresse chunk en cours de chargement
  LLint chunk_size;       // taille chunk en cours de chargement
  //
  char info[256];       // ventuel status pour le ftp
  int stop_ftp;         // flag stop pour ftp
#if DEBUG_CHECKINT
  char magic2;
#endif
} lien_back;

// cache
typedef struct {
  int version;        // 0 ou 1
  /* */
  int type;
  FILE *dat,*ndx,*olddat;
  char *use;      // liste des adr+fil
  FILE *lst;      // liste des fichiers pour la "purge"
  char lastmodified[256];
  // HASH
  void** hash;
  int hash_size;
  // fichiers log optionnels
  FILE* log;
  FILE* errlog;
  // variables
  int ptr_ant;      // pointeur pour anticiper
  int ptr_last;     // pointeur pour anticiper
} cache_back;

typedef struct {
  lien_url** liens;                     // pointeur sur liens
  int max_lien;                         // indice le plus grand rencontr
  int hash[3][HTS_HASH_SIZE];           // tables pour sav/adr-fil/former_adr-former_fil
} hash_struct;

#if HTS_HASH
#else
#define hash_write(A,B)
#endif

typedef struct {
  FILE* lst;
  char path[HTS_URLMAXSIZE*2];
} filecreate_params;


// Fonctions

// INCLUDES .H PARTIES DE CODE HTTRACK

// routine main
#include "htsmain.h"

// divers outils pour httrack.c
#include "htstools.h"

// aide pour la version en ligne de commande
#include "htshelp.h"

// gnration du nom de fichier  sauver
#include "htsname.h"

// gestion ftp
#include "htsftp.h"

// routine parser java
#include "htsjava.h"

// gestion interception d'URL
#include "htscatchurl.h"

// gestion robots.txt
#include "htsrobots.h"

// routines d'acceptation de liens
#include "htswizard.h"

// routines de regexp
#include "htsfilters.h"

// gestion backing
#include "htsback.h"

// gestion cache
#include "htscache.h"

// gestion hashage
#include "htshash.h"

// INCLUDES .H PARTIES DE CODE HTTRACK

//

// demande d'interaction avec le shell
#if HTS_ANALYSTE==2
//char HTbuff[1024];
void  hts_htmlcheck_init(void);
void  hts_htmlcheck_uninit(void);
int   hts_htmlcheck_start(void);
int   hts_htmlcheck_end(void);
int   hts_htmlcheck(char* html,int len,char* url_adresse,char* url_fichier);
char* hts_htmlcheck_query(char* question);
char* hts_htmlcheck_query2(char* question);
char* hts_htmlcheck_query3(char* question);
int   hts_htmlcheck_loop(lien_back* back,int back_max,int back_index,int lien_tot,int lien_ntot,LLint stat_bytes,LLint stat_bytes_recv,int stat_time,int stat_nsocket, LLint stat_written, int stat_updated, int stat_errors);
int   hts_htmlcheck_check(char* adr,char* fil,int status);
//
int hts_is_parsing(int flag);
int hts_is_testing(void);
int hts_setopt(httrackp* opt);
int hts_addurl(char** url);
int hts_resetaddurl(void);
int copy_htsopt(httrackp* from,httrackp* to);
char* hts_errmsg(void);
int hts_setpause(int);      // pause transfer
//
char* hts_cancel_file(char * s);
void hts_cancel_test(void);
void hts_cancel_parsing(void);
//
// Variables globales
extern int _hts_in_html_parsing;
extern int _hts_in_html_done;  // % raliss
extern int _hts_in_html_poll;  // parsing
extern char _hts_errmsg[1100];
extern int _hts_setpause;
extern httrackp* _hts_setopt;
extern char** _hts_addurl;
extern int _hts_cancel;
#endif



//


//int httpmirror(char* url,int level,httrackp opt);
int httpmirror(char* url1,int level1,httrackp opt);
int filesave(char* adr,int len,char* s);
void host_ban(httrackp* opt,lien_url** liens,int ptr,int lien_tot,lien_back* back,int back_max,char** filters,int filter_max,int* filptr,char* host);
FILE* filecreate(char* s);
int filenote(char* s,filecreate_params* params);
HTS_INLINE void usercommand(int exe,char* cmd,char* file);
void usercommand_exe(char* cmd,char* file);
int structcheck(char* s);
HTS_INLINE int fspc(FILE* fp,char* type);
char* next_token(char* p,int flag);
//
char* lirefichier(char* fil);
void check_rate(double stat_timestart,int maxrate);

// liens
int liens_record(char* adr,char* fil,char* save,char* former_adr,char* former_fil,char* codebase);


// backing, routines externes
int back_fill(lien_back* back,int back_max,httrackp* opt,cache_back* cache,lien_url** liens,int ptr,int numero_passe,int lien_tot);
int back_fillmax(lien_back* back,int back_max,httrackp* opt,cache_back* cache,lien_url** liens,int ptr,int numero_passe,int lien_tot);

// cancel file
#if HTS_ANALYSTE==2
char* hts_cancel_file(char * s);
void hts_cancel_test(void);
void hts_cancel_parsing(void);
#endif

int ask_continue(void);
int nombre_digit(int n);

// Java
int hts_add_file(char* file,int file_position);

// Polling
#if HTS_POLL
HTS_INLINE int check_flot(T_SOC s);
HTS_INLINE int check_stdin(void);
int read_stdin(char* s,int max);
#endif

void sig_finish( int code );     // finir et quitter
void sig_term( int code );       // quitter
#if HTS_WIN
void sig_ask( int code );        // demander
#else
void sig_back( int code );       // ignorer et mettre en backing 
void sig_ask( int code );        // demander
void sig_ignore( int code );     // ignorer signal
void sig_doback(int);            // mettre en arrire plan
#endif

// Void
void voidf(void);




#define HTS_TOPINDEX "TOP_INDEX_HTTRACK"

#define HTS_DATA_UNKNOWN_HTML "<html>"LF\
"<head>"LF\
"<title>Page not retrieved!</title>"LF\
"</head>"LF\
"<body background=\"backblue.gif\">"LF\
"<h1><strong><u>Oops!...</u></strong></h1>"LF\
"<h3>This page has <font color=\"red\"><em>not</em></font> been retrieved by HTTrack Website Copier. </h3>"LF\
"<script language=\"Javascript\">"LF\
"<!--"LF\
"  var loc=document.location.toString();"LF\
"  if (loc) {"LF\
"    var pos=loc.indexOf('link=');"LF\
"    if (pos>0) {"LF\
"      document.write('Clic to the link <b>below</b> to go to the online location!<br><a href=\"http://'+loc.substring(pos+5)+'\">http://'+loc.substring(pos+5)+'</a><br>');"LF\
"    } else"LF\
"      document.write('(no location defined)');"LF\
"  }"LF\
"// -->"LF\
"</script>"LF\
"<h6 align=\"right\">Mirror by HTTrack Website Copier</h6>"LF\
"</body>"LF\
"</html>"LF
#define HTS_DATA_UNKNOWN_HTML_LEN 0

#define HTS_DATA_ERROR_HTML "<html>"LF\
"<head>"LF\
"<title>Page not retrieved!</title>"LF\
"</head>"LF\
"<body background=\"backblue.gif\">"LF\
"<h1><strong><u>Oops!...</u></strong></h1>"LF\
"<h3>This page has <font color=\"red\"><em>not</em></font> been retrieved by HTTrack Website Copier (%s). </h3>"LF\
"<script language=\"Javascript\">"LF\
"<!--"LF\
"  var loc=document.location.toString();"LF\
"  if (loc) {"LF\
"    var pos=loc.indexOf('link=');"LF\
"    if (pos>0) {"LF\
"      document.write('Clic to the link <b>below</b> to go to the online location!<br><a href=\"http://'+loc.substring(pos+5)+'\">http://'+loc.substring(pos+5)+'</a><br>');"LF\
"    } else"LF\
"      document.write('(no location defined)');"LF\
"  }"LF\
"// -->"LF\
"</script>"LF\
"<h6 align=\"right\">Mirror by HTTrack Website Copier</h6>"LF\
"</body>"LF\
"</html>"LF

// image gif "unknown"
#define HTS_DATA_UNKNOWN_GIF "\x47\x49\x46\x38\x39\x61\x20\x0\x20\x0\xf7\xff\x0\xc0\xc0\xc0\xff\x0\x0\xfc\x3\x0\xf8\x6\x0\xf6\x9\x0\xf2\xc\x0\xf0\xf\x0\xf0\xe\x0\xed\x11\x0\xec\x13\x0\xeb\x14\x0\xe9\x15\x0\xe8\x18\x0\xe6\x18\x0\xe5\x1a\x0\xe3\x1c\x0\xe2\x1d\x0\xe1\x1e\x0\xdf\x20\x0\xdd\x23\x0\xdd\x22\x0\xdb\x23\x0\xda\x25\x0\xd9\x25\x0\xd8\x27\x0\xd6\x29\x0\xd5\x2a\x0\xd3\x2c\x0\xd2\x2d\x0"\
"\xd1\x2d\x0\xd0\x2f\x0\xcf\x30\x0\xce\x31\x0\xcb\x34\x0\xcb\x33\x0\xc8\x36\x0\xc5\x3b\x0\xc2\x3c\x0\xc0\x3f\x0\xbc\x43\x0\xba\x45\x0\xb7\x48\x0\xb4\x4c\x0\xb1\x4e\x0\xad\x51\x0\xaa\x55\x0\xa8\x58\x0\xa4\x5a\x0\xa1\x5e\x0\x9f\x60\x0\x99\x66\x0\x96\x68\x0\x93\x6c\x0\x90\x6e\x0\x8d\x72\x0\x8b\x74\x0\x8a\x75\x0\x88\x78\x0\x85\x79\x0\x82\x7d\x0\x7e\x80\x0\x7d\x82\x0\x79"\
"\x86\x0\x77\x88\x0\x73\x8b\x0\x72\x8d\x0\x70\x8e\x0\x6e\x91\x0\x6a\x95\x0\x68\x97\x0\x65\x9a\x0\x63\x9d\x0\x62\x9e\x0\x60\xa0\x0\x5d\xa2\x0\x5c\xa3\x0\x5a\xa5\x0\x57\xa9\x0\x57\xa7\x0\x54\xab\x0\x53\xac\x0\x52\xad\x0\x51\xae\x0\x4f\xb0\x0\x4e\xb1\x0\x4d\xb2\x0\x4c\xb4\x0\x49\xb6\x0\x48\xb8\x0\x46\xba\x0\x45\xbb\x0\x43\xbd\x0\x43\xbc\x0\x40\xbf\x0\x3f\xc0\x0\x3e\xc1"\
"\x0\x3d\xc2\x0\x3a\xc5\x0\x39\xc5\x0\x38\xc7\x0\x37\xc8\x0\x35\xca\x0\x34\xcb\x0\x32\xcc\x0\x31\xce\x0\x30\xd0\x0\x30\xce\x0\x2f\xd1\x0\x2e\xd1\x0\x2c\xd2\x0\x2b\xd4\x0\x2a\xd5\x0\x29\xd6\x0\x27\xd8\x0\x26\xda\x0\x26\xd8\x0\x25\xdb\x0\x24\xdc\x0\x21\xde\x0\x20\xdf\x0\x1f\xe1\x0\x1e\xe1\x0\x1c\xe3\x0\x1b\xe5\x0\x19\xe6\x0\x18\xe7\x0\x15\xeb\x0\x15\xea\x0\x14\xec\x0"\
"\x12\xed\x0\x10\xef\x0\xf\xf0\x0\xd\xf2\x0\xa\xf5\x0\x9\xf6\x0\x7\xf8\x0\x5\xfa\x0\x3\xfb\x0\x1\xfd\x0\x0\xfe\x2\x0\xfb\x4\x0\xf8\x7\x0\xf6\xa\x0\xf3\xd\x0\xee\x12\x0\xaa\x54\x0\xa5\x5a\x0\xa2\x5d\x0\xa0\x60\x0\x9c\x62\x0\x99\x66\x0\x98\x67\x0\x94\x6b\x0\x92\x6d\x0\x91\x6e\x0\x8f\x70\x0\x8c\x74\x0\x8a\x75\x0\x86\x79\x0\x83\x7c\x0\x81\x7e\x0\x7e\x82\x0"\
"\x7b\x83\x0\x79\x87\x0\x76\x8a\x0\x73\x8c\x0\x70\x8f\x0\x6a\x95\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0"\
"\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0"\
"\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x21\xf9\x4\x1\x0\x0\x0\x0\x2c\x0\x0\x0\x0\x20\x0\x20\x0\x40\x8"\
"\xff\x0\x1\x8\x1c\x48\xb0\x60\x82\x7\x16\x3a\x8c\x30\x91\x82\xc5\x8b\x82\x10\x23\xa\xa4\x81\x83\xa0\x92\x27\x56\xb6\x88\x51\x23\xb1\xa3\xc0\x38\x78\xfe\x10\x4a\xe4\xb1\xa4\xc9\x93\x1e\xf\x30\x90\x90\x41\xa2\x8e\x1e\x40\x88\x20\x41\x29\xf1\x4b\x99\x36\x75\xf6\xd0\x8c\xe8\x8\xd2\xce\x92\x94\x2e\x6d\xf2\x14\x8a\xd4\xcf\x92\x1\x4\x14\x58\x10\x1\xc3\x87\x82\x32\x6a\xe4\xe0\x81\x72\xc2\x86\x10"\
"\x1d\x83\x14\x49\xd2\x84\x8a\x96\xa3\x5\xa3\x5c\xe9\x42\x66\x8d\x1c\xb0\x4\xcf\xbc\xb1\xd3\x67\x10\x5a\x82\x80\xa\x29\x6a\xf4\xb6\xee\xce\x48\x93\x2c\x69\xb2\x4b\x50\x54\xa9\x53\x7c\x25\x6\x18\x60\xa0\x1\x5\xd\x22\x4a\xa0\x58\xe1\x22\xc6\xc\x1b\x34\x3\x10\x40\xe0\xa0\x2\x87\x88\x37\x76\xf8\x10\x82\x52\x1\x84\xb\x1e\x3a\xfe\x18\x62\x64\x9\x14\x94\x20\x48\x9c\x50\xd1\x2\x6\xc4\x23\x4c"\
"\xa4\x60\xf1\x12\xd8\xc9\x94\x2c\x60\xcc\xb8\xe1\x5b\x85\x4b\x18\x34\x70\xee\x4\x1e\x93\x66\x4e\x1e\x3f\x81\xd9\xd0\xd1\x13\xc8\x50\x60\x0\x7c\x4\x1d\x5a\xf4\x1c\x0\x22\x46\x8f\xaa\x6b\xdf\xce\xbd\xa3\xa4\x4a\x98\x38\x7d\xb\x7a\x9e\xa9\x13\xa8\x51\xa6\xba\xbf\xd\x8\x0\x3b\xff"
#define HTS_DATA_UNKNOWN_GIF_LEN 1070

#define HTS_DATA_BACK_GIF "\x47\x49\x46\x38\x39\x61\x40\x6\x4\x0\xf7\x0\x0\xb5\xc6\xff\xb5\xce\xff\xbd\xce\xff\xbd\xd6\xff\xc6\xd6\xff\xce\xd6\xff\xce\xde\xff\xd6\xde\xff\xd6\xe7\xff\xde\xe7\xff\xde\xef\xff\xe7\xef\xff\xef\xef\xff\xef\xf7\xff\xf7\xf7\xff\xf7\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"\
"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"\
"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"\
"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"\
"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"\
"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"\
"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"\
"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x21\xf9\x4\x1\x0\x0\x10\x0\x2c\x0\x0\x0\x0\x40\x6\x4\x0\x40\x8"\
"\xfe\x0\x1\x8\x1c\x48\x90\x60\x80\x83\x8\x13\x6\x10\xb0\x50\x80\xc3\x87\x10\x1d\x12\x10\x40\xa0\xa2\xc5\x8b\x18\xd\x10\xd0\x68\xa0\xa3\x47\x8f\x7\x42\x8a\x44\x70\x0\x1\x49\x93\x28\x4d\x26\x40\xb0\x32\x81\xcb\x97\x30\x5d\x2e\x48\x30\x73\x81\xcd\x9b\x38\x17\x30\xd0\xa9\x93\x81\x4f\x6\xd\x80\x2\x6d\xd0\xc0\x41\xd1\xa2\xe\x92\x2a\x5d\xca\xb4\xa9\xd3\xa7\x50\xa3\x4a\x9d\x4a\xb5\xaa\xd5\xab"\
"\x58\xb3\x6a\xdd\xca\xb5\xab\xd7\xaf\x60\xc3\x8a\x1d\x4b\xb6\xac\xd9\xb3\x68\xd3\xaa\x5d\xcb\xb6\xad\xdb\xb7\x70\xe3\xca\x9d\x4b\xb7\xae\xdd\xbb\x78\xf3\xea\xdd\xcb\x97\x6c\xc1\x81\xa\x1\x1c\x14\x2c\x0\x0\xc3\xc3\x11\x1f\x4e\x84\x38\x11\xe3\x45\x8e\x1b\xb\x44\xfe\x8\xd2\xc0\x1\xcb\x22\x33\x8b\x4c\x70\x80\xb3\xcb\xce\x31\x5f\xce\xa4\x49\x3a\x27\xce\x9d\x3c\x75\x6\x5d\x2d\xf4\xa8\x6b\xa3\x7d"\
"\x63\xcb\x9e\x4d\xbb\xb6\xed\xdb\xb8\x73\xeb\xde\xcd\xbb\xb7\xef\xdf\xc0\x83\xb\x1f\x4e\xbc\xf8\xdc\xbf\xc8\x7\x27\x4c\xcc\x5c\xa2\xe3\xc6\xcf\x39\x52\x9e\xde\xf1\x32\x2\x3\xd7\xb3\x97\x2c\x49\xb2\x65\xe8\xef\x35\x4d\x5f\x8b\xdf\xf9\xf3\xe7\x6a\xa2\x44\x19\xc0\x36\xce\xbe\xbd\xfb\xf7\xf0\xe3\xcb\x9f\x4f\xbf\xbe\xfd\xfb\xf8\xf3\xeb\xdf\xdf\x16\xb9\x60\xc0\x86\x19\x56\x58\x73\x14\x51\x54\x91\x81"\
"\xcf\x25\xb8\x51\x45\xd3\x15\x30\xdd\x65\x9a\x45\x18\x12\x67\xa0\x7d\x27\x5a\x69\xe1\x8d\xd7\x13\x79\xaa\xb5\x6\xd4\x7a\xfc\x85\x28\xe2\x88\x24\x96\x68\xe2\x89\x28\xa6\xa8\xe2\x8a\x2c\xb6\x38\x5c\x40\x0\x0\x3b\xff"
#define HTS_DATA_BACK_GIF_LEN 1153

#endif


